/*
** progadb3.c - utility to program the ADB3 bridge firmware.
**
** (C) Copyright 2009-2013 Alpha Data
**
** Notes:
**
**  1. ADM-XRC-6TL
**     ADM-XRC-6T1
**     ADM-XRC-6TGE(L)
**     ADPE-XRC-6T(-L)
**     ADM-XRC-6T-DA1
**     ADPE-XRC-6T-ADV (Controller)
**     ADM-XRC-7K1
**     ADM-XRC-7V1
**     ADM-VPX3-7V2
**
**     There are two regions in Flash bank 0 used to configure the bridge
**     FPGA (byte addresses):
**
**        Failsafe        0x0 - 0x7FFFFF
**        Normal     0x800000 - 0xFFFFFF
**
**     If the bridge FPGA cannot be configured succesfully from the Normal
**     image, the onboard firmware falls back to the Failsafe image. Thus,
**     the default behaviour of this utility is to write the Normal image.
**
**     To write the Failsafe image, the option '+failsafe' must be passed,
**     although the Failsafe image should NEVER be written except on advice
**     of Alpha Data.
**
**     The full Flash bank 0 byte-address map for these models is:
**
**     0x0000000 - 0x07FFFFF  Failsafe ADB3 PCI-E to OCP Bridge bitstream (*)
**     0x0800000 - 0x0FFFFFF  Normal ADB3 PCI-E to OCP Bridge bitstream (*)
**     0x1000000 - 0x10FFFFF  Alpha Data VPD region
**     0x1100000 - 0x11FFFFF  Customer VPD region
**     0x1200000 - 0x28FFFFF  Target FPGA bitstream
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA bitstream
**
**     This utility, as supplied by Alpha Data, writes only to the regions
**     marked (*), and does not affect the other regions.
**
**  2. ADPE-XRC-6T-ADV (Controller)
**
**     There are two regions in Flash bank 0 used to configure the Controller
**     FPGA (byte addresses):
**
**        Failsafe        0x0 - 0x7FFFFF
**        Normal     0x800000 - 0xFFFFFF
**
**     If the bridge FPGA cannot be configured succesfully from the Normal
**     image, the onboard firmware falls back to the Failsafe image. Thus,
**     the default behaviour of this utility is to write the Normal image.
**
**     To write the Failsafe image, the option '+failsafe' must be passed,
**     although the Failsafe image should NEVER be written except on advice
**     of Alpha Data.
**
**     The full Flash address map for the ADPE-XRC-6T-ADV (Controller) is:
**
**     Flash bank 0 (U52):
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream (*)
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream (*)
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream
**
**     Flash bank 1 (U51):
**     0x0000000 - 0x0FFFFFF  Customer region
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 1 bitstream
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 1 bitstream
**
**  3. ADM-XRC-KU1
**
**     There are two regions in Flash bank 0 used to configure the bridge
**     FPGA (byte addresses):
**
**        Failsafe        0x0 - 0x7FFFFF
**        Normal     0x800000 - 0xFFFFFF
**
**     If the bridge FPGA cannot be configured succesfully from the Normal
**     image, the onboard firmware falls back to the Failsafe image. Thus,
**     the default behaviour of this utility is to write the Normal image.
**
**     To write the Failsafe image, the option '+failsafe' must be passed,
**     although the Failsafe image should NEVER be written except on advice
**     of Alpha Data.
**
**     The full Flash bank 0 byte-address map for these models is:
**
**     0x0000000 - 0x07FFFFF  Failsafe ADB3 PCI-E to OCP Bridge bitstream (*)
**     0x0800000 - 0x0FFFFFF  Normal ADB3 PCI-E to OCP Bridge bitstream (*)
**     0x1000000 - 0x10FFFFF  Alpha Data VPD region
**     0x1100000 - 0x11FFFFF  Customer VPD region
**     0x1200000 - 0x48FFFFF  Target FPGA bitstream
**     0x4900000 - 0x7FFFFFF  Alternate target FPGA bitstream
**
** NOTE: This utility, as supplied by Alpha Data, writes only to the regions
**       marked (*), and does not affect the other regions.
*/

#if defined(_WIN32) || defined(_MSC_VER)

/* MSVC or DDK */
# include <windows.h>
# include <tchar.h>
typedef BOOL boolean_t;

#else

/* Other compiler */
# define _T(x) x
# if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
    /* VxWorks 5.x doesn't have snprintf() */
#   define _stprintf_s(buf, max, format, ...) sprintf(buf, format, __VA_ARGS__)
# else
#   define _stprintf_s(buf, max, format, ...) snprintf(buf, max, format, __VA_ARGS__)
# endif
# define _tcslen(x) strlen(x)
# define _totlower(x) tolower(x)
# define _tmain main
# define _tprintf printf
# define FALSE (0)
# define TRUE (1)
typedef int boolean_t;

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <test_common.h>
#include <platform_endian.h>
#if defined(__VXWORKS__) || defined(__vxworks)
# include <platform_main.h>
#else
# define DEFINE_MAIN(prog) int _tmain(int argc, TCHAR* argv[])
# define PROGRAM_NAME(prog)
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define FLCTL  0x0040
#define FLPAGE 0x0044
#define FLDATA 0x0048

static const uint64_t maxError = 10U;

static ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;

static void* pCtlBase;
static volatile uint8_t* pCtlBase8;
static volatile uint16_t* pCtlBase16;
static volatile uint32_t* pCtlBase32;

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

static bool_t
checkFileExtension(
  const TCHAR* pFilename,
  const TCHAR* pExtension)
{
  size_t m = _tcslen(pFilename);
  size_t n = _tcslen(pExtension);
  TCHAR c, d;
  size_t i;

  if (m < n) {
    return FALSE;
  }
  pFilename += m - n;
  i = 0;
  do {
    c = _totlower(*pExtension);
    pExtension++;
    if (c != _T('\0')) {
      d = *pFilename++;
      if (_totlower(d) != c) {
        /* Extension not found */
        return FALSE;
      }
    }
  } while (c != _T('\0') && i < n);

  /* Found the file extension */
  return TRUE;
}

static ADMXRC3_STATUS
flash_write(
  uint32_t location,
  uint16_t value)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint32_t flpage, flctl;

  /* Set the page */
  flpage = location & 0x3fe0000U;
  pCtlBase32[FLPAGE >> 2] = cpu_to_le32(flpage);
  (void)pCtlBase32[FLPAGE >> 2];
  /* Set the data */
  pCtlBase16[FLDATA >> 1] = cpu_to_le16(value);
  (void)pCtlBase16[FLDATA >> 1];
  /*
  ** Select mode 0, set fl_a, Drive fl_a, fl_d, fl_ce_n, fl_oe_n, 
  ** fl_we_n, assert fl_ce_n
  */
  flctl = (location & 0x1fffeU) | (0x7U << 24) | (0x1U << 28);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Assert fl_we_n */
  flctl |= (0x1U << 30);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Deassert fl_we_n */
  flctl &= ~(0x1U << 30);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Deassert fl_ce_n */
  flctl &= ~(0x1U << 28);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Stop driving fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n */
  flctl &= ~(0x7U << 24);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];

  return (status);   
}

static uint16_t
flash_read(
  unsigned int location)
{
  uint32_t flpage, flctl;
  uint16_t val16;

  /* Set the page */
  flpage = location & 0x3fe0000U;
  pCtlBase32[FLPAGE >> 2] = cpu_to_le32(flpage);
  (void)pCtlBase32[FLPAGE >> 2];
  /* 
  ** Select mode 0, set fl_a, Drive fl_a, fl_ce_n, fl_oe_n, 
  ** fl_we_n, assert fl_ce_n
  */
  flctl = (location & 0x1fffeU) | (0x3U << 25) | (0x1U << 28);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Assert fl_oe_n */
  flctl |= (0x1U << 29);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Read fl_d */
  val16 = pCtlBase16[FLDATA >> 1];
  val16 = le16_to_cpu(val16);
  /* Deassert fl_oe_n */
  flctl &= ~(0x1U << 29);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Deassert fl_ce_n */
  flctl &= ~(0x1U << 28);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  /* Stop driving fl_a, fl_ce_n, fl_oe_n, fl_we_n */
  flctl &= ~(0x3U << 25);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
  
  return val16;
}

static void
init_flash_interface(
  void)
{
  uint32_t flctl;

  /*
  * reset the flash memory
  */
  flctl = (0x88000000U);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];

  flctl &= ~(0x1U << 31);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];

  flctl &= ~(0x1U << 27);
  pCtlBase32[FLCTL >> 2] = cpu_to_le32(flctl);
  (void)pCtlBase32[FLCTL >> 2];
}

static ADMXRC3_STATUS
flash_block_erase(
  uint32_t location)
{
  ADMXRC3_STATUS status;
  uint16_t val16;
  uint32_t timeout;

  flash_write(location, 0x50U); /* Clear status */
  flash_write(location, 0xFFU); /* Read array */
  flash_write(location, 0x60U); /* Block unlock */
  flash_write(location, 0xD0U); /* Confirm block unlock */
  flash_write(location, 0x20U); /* Block erase */
  flash_write(location, 0xD0U); /* Confirm block erase */
  /* Wait for ready */
  timeout = 0;
  val16 = flash_read(location);
  while (!(val16 & 0x0080U)) {
    val16 = flash_read(location);
    timeout++;
    if (timeout > 1000000) {
      break;
    }
  }
  if (val16 != 0x0080U) {
    status = ADMXRC3_HARDWARE_ERROR;
    _tprintf(_T("Block erase failed at 0x%08lx, status 0x%02x\n"),
      (unsigned long)location, (unsigned int)val16);
  } else {
    status = ADMXRC3_SUCCESS;
  }
  flash_write(location, 0x50U); /* Clear status */
  flash_write(location, 0xFFU); /* Read array */

  return status;
}

static ADMXRC3_STATUS
flash_block_program(
  size_t bufferSize,
  uint32_t blockAddr,
  uint8_t* p,
  uint32_t address,
  size_t n)
{
  uint16_t val16;
  size_t nWord;

  n >>= 1;
  while (n) {
    /* Program-buffer command */
    flash_write(blockAddr, 0xE8U);
    /* Wait for ready */
    do {
      val16 = flash_read(blockAddr);
    } while ((val16 & 0x80U) != 0x80U);
    /* Calculate number of words to program; won't cross a buffer address boundary */
    nWord = bufferSize - (size_t)((address >> 1) & (bufferSize - 1));
    if (nWord > n) {
      nWord = n;
    }
    /* Number of words */
    flash_write(blockAddr, (uint16_t)(nWord - 1U));
    /* Write data */
    n -= nWord;
    while (nWord != 0) {
      val16  = (uint16_t)((uint16_t)*p++ << 8);
      val16 = (uint16_t)(val16 | ((uint16_t)*p++ << 0));
      flash_write((uint32_t)address, val16);
      address += 2;
      nWord--;
    }
    /* Confirm buffer-write */
    flash_write(blockAddr, 0xD0U);
    /* Wait for completion */
    do {
      val16 = flash_read(blockAddr);
    } while ((val16 & 0x80U) != 0x80U);
    /* Clear status */
    flash_write(blockAddr, 0x50U);
    /* Read array command */
    flash_write(blockAddr, 0xFFU);
    /* Check status */
    if ((val16 & 0x7E) != 0) {
      _tprintf(_T("*** Buffered write failed, Flash status 0x%02x\n"), (unsigned int)val16);
      return ADMXRC3_HARDWARE_ERROR;
    }
  }

  return ADMXRC3_SUCCESS;
}

static void
verify_bitstream(
  uint8_t* p,
  uint32_t address,
  size_t size,
  uint64_t* pnErr)
{
  uint16_t val16;
  uint8_t exp;
  uint8_t val8;
  size_t nErr = 0;

  flash_write(address, 0xFFU); /* Read array comand */
  while (size) {
    val16 = flash_read(address);
    val8  = (uint8_t)((val16 >> 8) & 0x00FFU);
    exp   = *p++;
    if (exp != val8) {
      if (nErr == maxError) {
        _tprintf(_T("*** expected 0x%02lx actual 0x%02lx index 0x%lx\n"),
          (unsigned long)exp, (unsigned long)val8, (unsigned long)address);
      } else {
        _tprintf(_T("*** more errors\n"));
      }
      nErr++;
    }
    size--;
    if (size) {
      val8 = (uint8_t)(val16  & 0x00FFU);
      exp = *p++;
      if (exp != val8) {
        if (nErr == maxError) {
          _tprintf(_T("*** expected 0x%02lx actual 0x%02lx index 0x%lx\n"),
            (unsigned long)exp, (unsigned long)val8, (unsigned long)address);
        } else {
          _tprintf(_T("*** more errors\n"));
        }
        nErr++;
      }
      size--;
    }
    address += 2;
  }
  *pnErr = *pnErr + nErr;
}

static ADMXRC3_STATUS load_bitstream(
  ADMXRC3_HANDLE hCard,
  const TCHAR* filename,
  uint8_t** ppBitstream,
  size_t* image_size,
  boolean_t bBitReversal)
{
  ADMXRC3_CARD_INFO info;
  ADMXRC3_STATUS status;
  FILE* file;
  uint8_t* pBitstream = NULL;
  long fileSize;
  size_t numRead;

  status = ADMXRC3_GetCardInfo(hCard, &info);
  if (status != ADMXRC3_SUCCESS) {
    goto done;
  }

  if (_tfopen_s(&file, filename, _T("rb")) != 0) {
    status = ADMXRC3_FILE_NOT_FOUND;
    goto done;
  }

  if (fseek(file, 0, SEEK_END) != 0) {
    status = ADMXRC3_INVALID_BITSTREAM;
    goto done;
  }

  fileSize = ftell(file);
  if (fileSize < 0) {
    status = ADMXRC3_INVALID_BITSTREAM;
    goto done;
  }

  rewind(file);

  pBitstream = (uint8_t*)malloc((size_t)fileSize); /* Cast of long >= 0 to size_t should be OK */
  if (pBitstream == NULL) {
    status = ADMXRC3_NO_MEMORY;
    fclose(file);
    goto done;
  }

  fseek(file, 0, 0);
  numRead = fread(pBitstream, 1, (size_t)fileSize /* Cast of long >= 0 to size_t should be OK */, file);
  if (numRead != fileSize) {
    status = ADMXRC3_INVALID_BITSTREAM;
    free(pBitstream);
    fclose(file);
    goto done;
  }

  fclose(file);

  if (bBitReversal) {
    long i;
    uint8_t val8;

    for (i = 0; i < fileSize; i++) {
      val8 = pBitstream[i];
      val8 =
        ((val8 & 0x01U) << 7) | ((val8 & 0x02U) << 5) | ((val8 & 0x04U) << 3) | ((val8 & 0x08U) << 1) |
        ((val8 & 0x10U) >> 1) | ((val8 & 0x20U) >> 3) | ((val8 & 0x40U) >> 5) | ((val8 & 0x80U) >> 7);
      pBitstream[i] = val8;
    }
  }

  *ppBitstream = pBitstream;
  *image_size = fileSize;

done:
  return status;
}  /* end load_bitstream */

DEFINE_MAIN(adb3ProdProgAdb3)
{
  int ret = 0;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  uint8_t* pBitstream = NULL;
  Arguments arguments = g_initArguments;
  Option options[3];
  OptionType paramType[1];
  TCHAR* paramInfo[1];
  TCHAR* paramMeaning[1];
  boolean_t bGetBuildDate, bBitReversal = FALSE;
  uint64_t imageAddr;
  uint64_t eraseAddr, eraseLimit;
  uint64_t location;
  uint64_t nErr;
  size_t imageSize, remaining, bufferSize;
  TCHAR* filename;
  unsigned int buildDateLo, buildDateHi;
  TCHAR buildDateStr[64];
  boolean_t bFailsafeImage = FALSE;
  unsigned int index, serial, windowIndex;
  ADMXRC3_FLASH_INFO flashInfo;
  ADMXRC3_FLASHBLOCK_INFO flashBlockInfo;
  uint8_t* p;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("failsafe");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("program failsafe image");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  paramType[0] = OptionString;
  paramInfo[0] = _T("<.BIN file>");
  paramMeaning[0] = _T("specify binary bitstream file (.BIN)");
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;

  ret = sampleParseCommandLine(_T("progadb3"), argc, argv, &arguments);
  if (ret < 0) {
    goto done;
  }

  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  filename = arguments.param[0].pStringVal;
  bFailsafeImage = options[2].value.booleanVal;

  if (!checkFileExtension(filename, _T(".bin"))) {
    _tprintf(_T("*** Specified file does not have .bin file extension\n"));
    ret = -1;
    goto done;
  }

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -2;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMVPX37V2:
    if (bFailsafeImage) {
      imageAddr    = 0x00000000U;
      eraseAddr    = 0x00000000U;
      eraseLimit   = 0x007FFFFFU;
    } else {
      imageAddr    = 0x00800000U;
      eraseAddr    = 0x00800000U;
      eraseLimit   = 0x00FFFFFFU;
    }
    bufferSize = 512;
    windowIndex = 2;
    break;

  default:
    _tprintf(_T("*** This utility must be run on one of the following models:\n\n"));
    _tprintf(_T("\to ADM-XRC-6TL\n"));
    _tprintf(_T("\to ADM-XRC-6T1\n"));
    _tprintf(_T("\to ADM-XRC-6T-DA1\n"));
    _tprintf(_T("\to ADM-XRC-6TGE(L)\n"));
    _tprintf(_T("\to ADPE-XRC-6T(-L)\n"));
    _tprintf(_T("\to ADPE-XRC-6T-ADV (Controller)\n"));
    _tprintf(_T("\to ADM-XRC-7K1\n"));
    _tprintf(_T("\to ADM-XRC-7V1\n"));
    _tprintf(_T("\to ADM-VPX3-7V2\n"));
    _tprintf(_T("\to ADM-XRC-KU1\n"));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_MapWindow(hCard, windowIndex, 0, 0x1000, &pCtlBase);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to map window: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  pCtlBase8 = (volatile uint8_t*)pCtlBase;
  pCtlBase16 = (volatile uint16_t*)pCtlBase;
  pCtlBase32 = (volatile uint32_t*)pCtlBase;

  status = ADMXRC3_GetFlashInfo(hCard, 0, &flashInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get Flash information: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
  case ADMXRC3_MODEL_ADMVPX37V2:
    bGetBuildDate = TRUE;
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
    bGetBuildDate = TRUE;
    bBitReversal = TRUE;
    break;

  default:
    bGetBuildDate = FALSE; /* should never get here anyway */
    break;
  }

  if (bGetBuildDate) {
    buildDateLo = le32_to_cpu(pCtlBase32[3]);
    buildDateHi = le32_to_cpu(pCtlBase32[2]);
    _stprintf_s(buildDateStr, sizeof(buildDateStr) / sizeof(buildDateStr[0]), _T("%02x/%02x/%04x %02x:%02x:%02x"),
      (unsigned int)((buildDateHi >> 24) & 0xffU),
      (unsigned int)((buildDateHi >> 16) & 0xffU),
      (unsigned int)((buildDateHi >> 0) & 0xffffU),
      (unsigned int)((buildDateLo >> 24) & 0xffU),
      (unsigned int)((buildDateLo >> 16) & 0xffU),
      (unsigned int)((buildDateLo >> 8) & 0xffU));
    _tprintf(_T("Current bridge build date: %s (0x%8.8lx_%8.8lx)\n"),
      buildDateStr, (unsigned long)buildDateHi, (unsigned long)buildDateLo);
  } else {
    _tprintf(_T("Current bridge build date: not available (Bridge revision too low)\n"));
  }

  status = ADMXRC3_GetFlashInfo(hCard, 0, &flashInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get Flash information: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  _tprintf(_T("Flash type is %s, %llu(0x%llx) kiB\n"), flashInfo.Identifier,
    (unsigned long long)(flashInfo.Size / 1024), (unsigned long long)(flashInfo.Size / 1024));

  status = load_bitstream(hCard, filename, &pBitstream, &imageSize, bBitReversal);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to load bitstream %s: %s\n"), filename, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  /*
  **  Initialise the device, check ID
  */
  init_flash_interface();

  _tprintf(_T("Erasing Flash 0x%lx - 0x%lx...\n"), (unsigned long)eraseAddr, (unsigned long)eraseLimit);
  location = eraseAddr;
  while (location < eraseLimit) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, location, &flashBlockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block information for location 0x%llx: %s\n"),
        (unsigned long long)location, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    _tprintf(_T("0x%llx\r"), (unsigned long long)location);
    fflush(stdout);
    flash_block_erase((uint32_t)location);
    location += flashBlockInfo.Length;
  }

  _tprintf(_T("Programming Flash 0x%lx - 0x%lx...\n"),
    (unsigned long)imageAddr, (unsigned long)(imageAddr + imageSize - 1));
  location = imageAddr;
  remaining = imageSize;
  p = (uint8_t*)pBitstream;
  while (remaining != 0) {
    size_t chunk;

    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, location, &flashBlockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block information for location 0x%llx: %s\n"),
        (unsigned long long)location, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    _tprintf(_T("0x%llx\r"), (unsigned long long)location);
    fflush(stdout);

    chunk = (size_t)(flashBlockInfo.Address + flashBlockInfo.Length - location);
    if (chunk > remaining) {
      chunk = remaining;
    }
    flash_block_program(bufferSize, (uint32_t)flashBlockInfo.Address, p, (uint32_t)location, chunk);
    p += chunk;
    remaining -= chunk;
    location += chunk;
  }

  _tprintf(_T("Verifying Flash 0x%lx - 0x%lx...\n"),
    (unsigned long)imageAddr, (unsigned long)(imageAddr + imageSize - 1));
  location = imageAddr;
  remaining = imageSize;
  p = (uint8_t*)pBitstream;
  nErr = 0;
  while (remaining != 0) {
    size_t chunk;

    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, location, &flashBlockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block information for location 0x%llx: %s\n"),
        (unsigned long long)location, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    _tprintf(_T("0x%llx\r"), (unsigned long long)location);
    fflush(stdout);

    chunk = (size_t)(flashBlockInfo.Address + flashBlockInfo.Length - location);
    if (chunk > remaining) {
      chunk = remaining;
    }
    verify_bitstream(p, (uint32_t)location, chunk, &nErr);
    p += chunk;
    remaining -= chunk;
    location += chunk;
  }

  if (nErr) {
    _tprintf(_T("*** FAILED with 0x%llx data error(s)\n"), (unsigned long long)nErr);
    ret = -1;
    goto done;
  } else {
    _tprintf(_T("PASSED             \n"));
  }

done: 
  if (pBitstream != NULL) {
    free(pBitstream);
    pBitstream = NULL;
  }
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

PROGRAM_NAME(adb3ProdProgAdb3)
